#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['idf_build_apps', 'idf_build_apps.manifest']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing', 'pyyaml', 'packaging']

extras_require = \
{":python_version < '3.4'": ['pathlib'],
 'doc': ['sphinx',
         'sphinx-rtd-theme',
         'sphinx_copybutton',
         'myst-parser',
         'sphinxcontrib-mermaid'],
 'test': ['pytest', 'pytest-cov']}

setup(name='idf-build-apps',
      version='0.4.1',
      description='Tools for building ESP-IDF related apps.',
      author=None,
      author_email='Fu Hanxi <fuhanxi@espressif.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
     )
